from pathlib import Path
from datetime import timezone

from mxbiflow import set_base_path
from mxbiflow.bootstrap import init_gameloop
from mxbiflow.core.path import get_email_state_path, get_mxbi_config_path

from mxbiflow.infra.post_processing import HtmlComposer, session_overview, summarize
from mxbiflow.models.session import EmailSendStateStore
from mxbiflow.scene import SceneManager
from mxbiflow.scene.idle.idle import IDLE
from mxbiflow.ui.wizard import config_wizard
from pymotego.email import EmailClient
from mxbiflow.core.config_store import ConfigStore
from pymxbi.mxbi.factory import MXBIModel


def main() -> None:
    set_base_path(Path.cwd())

    scene_manager = SceneManager()
    scene_manager.register([IDLE])

    config_wizard(scene_manager)

    game = init_gameloop(scene_manager)
    game.play()

    summary = summarize()

    date_str = (
        summary.start_at.astimezone(timezone.utc).strftime("%Y-%m-%d")
        if summary.start_at
        else "N/A"
    )
    composer = HtmlComposer(
        title=f"Session #{summary.session_id}",
        date=date_str,
    )
    composer.add_section(session_overview())

    html = composer.html

    mxbi_config = ConfigStore(get_mxbi_config_path(), MXBIModel).value
    email_store = EmailSendStateStore(get_email_state_path())
    prev_state = email_store.load()

    with EmailClient() as client:
        result = client.send(
            subject=f"{mxbi_config.mxbi_id} Daily Report",
            html_body=html,
            in_reply_to=prev_state.message_id if prev_state.message_id else None,
        )
        email_store.save(result.message_id)


if __name__ == "__main__":
    main()



import os
import tempfile
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path

from pydantic import BaseModel, Field, PrivateAttr, ValidationError, field_serializer

from .animal import Animal, AnimalConfig
from .reward import RewardEnum


class SessionConfig(BaseModel):
    experimenter: str = Field(default="auto", frozen=True)
    reward_type: RewardEnum = Field(default=RewardEnum.AGUM_ONE_FIFTH, frozen=True)
    send_email: bool = Field(default=False, frozen=True)
    sync_data: bool = Field(default=False, frozen=True)
    note: str = Field(default="", max_length=1000)

    default_scene: str = Field(default="")
    unknown_animal_fallback: str = Field(default="")
    fault_fallback: str = Field(default="")

    hide_cursor: bool = Field(default=False)
    fullscreen: bool = Field(default=False)

    auto_accept_timeout_seconds: int = Field(default=60, ge=0)

    animals: list[AnimalConfig] = Field(default_factory=list)


class DailySessionCounter(BaseModel):
    day: str = Field(default_factory=lambda: datetime.now().date().isoformat())
    last_session_id: int = Field(default=0, ge=0)


class EmailSendState(BaseModel):
    sent_at: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))
    message_id: str = Field(default="")


@dataclass
class DailySessionIdStore:
    path: Path

    def _today_local(self) -> str:
        return datetime.now().date().isoformat()

    def _load(self) -> DailySessionCounter:
        if not self.path.exists():
            return DailySessionCounter()

        text = self.path.read_text()

        try:
            return DailySessionCounter.model_validate_json(text)
        except ValueError, ValidationError:
            return DailySessionCounter()

    def _atomic_write(self, data: DailySessionCounter) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        fd, tmp = tempfile.mkstemp(
            dir=str(self.path.parent),
            prefix=self.path.name,
            suffix=".tmp",
        )

        try:
            with os.fdopen(fd, "w", encoding="utf-8") as f:
                f.write(data.model_dump_json())
                f.flush()
                os.fsync(f.fileno())

            os.replace(tmp, self.path)
        finally:
            try:
                os.remove(tmp)
            except FileNotFoundError:
                pass

    @property
    def session_id(self) -> int:
        today = self._today_local()
        data = self._load()

        if data.day != today:
            data.day = today
            data.last_session_id = 0

        data.last_session_id += 1
        self._atomic_write(data)

        return data.last_session_id


@dataclass
class EmailSendStateStore:
    path: Path

    def _load(self) -> EmailSendState:
        if not self.path.exists():
            return EmailSendState()

        text = self.path.read_text()

        try:
            return EmailSendState.model_validate_json(text)
        except ValueError, ValidationError:
            return EmailSendState()

    def _atomic_write(self, data: EmailSendState) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        fd, tmp = tempfile.mkstemp(
            dir=str(self.path.parent),
            prefix=self.path.name,
            suffix=".tmp",
        )

        try:
            with os.fdopen(fd, "w", encoding="utf-8") as f:
                f.write(data.model_dump_json())
                f.flush()
                os.fsync(f.fileno())

            os.replace(tmp, self.path)
        finally:
            try:
                os.remove(tmp)
            except FileNotFoundError:
                pass

    def load(self) -> EmailSendState:
        return self._load()

    def save(self, message_id: str) -> None:
        data = EmailSendState(message_id=message_id)
        self._atomic_write(data)


class Session(BaseModel):
    experimenter: str = Field(default="auto", frozen=True)
    reward_type: RewardEnum = Field(default=RewardEnum.AGUM_ONE_FIFTH, frozen=True)
    send_email: bool = Field(default=False, frozen=True)
    sync_data: bool = Field(default=False, frozen=True)

    default_scene: str = Field(default="", frozen=True)
    unknown_animal_fallback: str = Field(default="", frozen=True)
    fault_fallback: str = Field(default="", frozen=True)

    hide_cursor: bool = Field(default=False, frozen=True)
    fullscreen: bool = Field(default=False, frozen=True)

    session_id: int = Field(default=0, ge=0)
    start_at: datetime | None = Field(default=None)
    end_at: datetime | None = Field(default=None)
    note: str = Field(frozen=True)

    _current_animal: str | None = PrivateAttr(default=None)
    _session_store: DailySessionIdStore | None = PrivateAttr(default=None)
    animals: dict[str, Animal] = Field(default_factory=dict, frozen=True)

    def set_session_store(self, store: DailySessionIdStore) -> None:
        self._session_store = store

    def start(self):
        if self._session_store and self.session_id == 0:
            self.session_id = self._session_store.session_id
        self.start_at = datetime.now(timezone.utc)

    def end(self):
        self.end_at = datetime.now(timezone.utc)

    @field_serializer("start_at", "end_at", when_used="json")
    def _serialize_timestamp(self, value: datetime | None) -> str | None:
        if value is None:
            return None

        return value.astimezone().isoformat(timespec="microseconds")

    @property
    def current_animal(self) -> Animal | None:
        key = self._current_animal
        if key is None:
            return None

        return self.animals[key]

    @property
    def require_current_animal(self) -> Animal:
        key = self._current_animal
        if key is None:
            raise RuntimeError(
                f"Animal: {key} is not set. Please call set_current_animal() first"
            )

        return self.animals[key]

    def clear_current_animal(self):
        a = self.current_animal
        if a is None:
            return

        a.end_animal_session()
        self._current_animal = None

    def set_current_animal(self, animal: str):
        self.clear_current_animal()

        try:
            a = self.animals[animal]
        except KeyError:
            raise ValueError(f"animal {animal} not found")

        self._current_animal = animal
        a.start_animal_session()


class Options(BaseModel):
    mxbis: list[str] = Field(default_factory=list, frozen=True)
    experimenter: list[str] = Field(default_factory=list, frozen=True)
    animals: dict[str, str] = Field(default_factory=dict, frozen=True)



import pygame
from pygame import Surface


def create_background(screen_size: tuple[int, int], border_width: int = 40) -> Surface:
    width, height = screen_size
    surface = Surface(screen_size)
    surface.fill((255, 255, 255))
    black_rect = pygame.Rect(
        border_width, border_width, width - 2 * border_width, height - 2 * border_width
    )
    pygame.draw.rect(surface, (0, 0, 0), black_rect)
    return surface



from datetime import datetime
from pathlib import Path
from typing import Self, TypedDict

from jinja2 import Environment, FileSystemLoader, select_autoescape
from pydantic import BaseModel

from ..core.context import get_mxbiflow


class StageSummary(BaseModel):
    name: str
    trials: int
    initial_level: int
    final_level: int


class AnimalSummary(BaseModel):
    name: str
    rfid_id: str
    total_trials: int
    animal_sessions: int
    total_duration_seconds: float
    stages: list[StageSummary]


class SessionSummary(BaseModel):
    session_id: int
    start_at: datetime | None
    end_at: datetime | None
    duration_seconds: float
    reward_type: str
    total_animals: int
    animals: list[AnimalSummary]


class ReportImage(TypedDict):
    cid: str
    alt: str


def _format_timestamp(dt: datetime | None) -> str:
    if dt is None:
        return "N/A"

    return dt.astimezone().strftime("%Y-%m-%d %H:%M:%S")


def _format_duration(seconds: float) -> str:
    if seconds <= 0:
        return "N/A"
    hours, remainder = divmod(int(seconds), 3600)
    minutes, secs = divmod(remainder, 60)
    if hours > 0:
        return f"{hours}h {minutes}m {secs}s"
    elif minutes > 0:
        return f"{minutes}m {secs}s"
    return f"{secs}s"


def _calc_animal_duration(sessions: list) -> float:
    total = 0.0
    for s in sessions:
        if s.start_at and s.end_at:
            total += (s.end_at - s.start_at).total_seconds()
    return total


_TEMPLATE_DIR = Path(__file__).parent / "templates"


def _get_jinja_env() -> Environment:
    return Environment(
        loader=FileSystemLoader(_TEMPLATE_DIR),
        autoescape=select_autoescape(["html", "xml"]),
    )


def summarize() -> SessionSummary:
    session = get_mxbiflow().session
    animal_summaries: list[AnimalSummary] = []

    for name, animal in session.animals.items():
        stages: list[StageSummary] = []
        for stage_name, stage_state in animal._stages.items():
            stages.append(
                StageSummary(
                    name=stage_name,
                    trials=stage_state.stage_trial_id,
                    initial_level=stage_state.initial_level,
                    final_level=stage_state.level,
                )
            )

        animal_summaries.append(
            AnimalSummary(
                name=name,
                rfid_id=animal.rfid_id,
                total_trials=animal.trial_id,
                animal_sessions=len(animal._sessions),
                total_duration_seconds=_calc_animal_duration(animal._sessions),
                stages=stages,
            )
        )

    duration = 0.0
    if session.start_at and session.end_at:
        duration = (session.end_at - session.start_at).total_seconds()

    return SessionSummary(
        session_id=session.session_id,
        start_at=session.start_at,
        end_at=session.end_at,
        duration_seconds=duration,
        reward_type=session.reward_type.value,
        total_animals=len(session.animals),
        animals=animal_summaries,
    )


def session_overview() -> str:
    summary = summarize()
    env = _get_jinja_env()
    template = env.get_template("session_overview.html")

    context = {
        "session_id": summary.session_id,
        "session_date": (
            summary.start_at.astimezone().strftime("%Y-%m-%d")
            if summary.start_at
            else "N/A"
        ),
        "start_time": _format_timestamp(summary.start_at),
        "end_time": _format_timestamp(summary.end_at),
        "duration": _format_duration(summary.duration_seconds),
        "reward_type": summary.reward_type,
        "total_animals": summary.total_animals,
        "animals": [
            {
                **a.model_dump(),
                "duration": _format_duration(a.total_duration_seconds),
            }
            for a in summary.animals
        ],
    }

    return template.render(**context)


def render_image_section_report(section_title: str, images: list[ReportImage]) -> str:
    env = _get_jinja_env()
    template = env.get_template("section_report.html")
    return template.render(section_title=section_title, images=images)


class HtmlComposer:
    def __init__(self, title: str, date: str) -> None:
        self._title = title
        self._date = date
        self._sections: list[str] = []
        self._env = _get_jinja_env()

    def add_section(self, html: str) -> Self:
        self._sections.append(html)
        return self

    @property
    def html(self) -> str:
        template = self._env.get_template("page_layout.html")
        return template.render(
            title=self._title,
            date=self._date,
            sections=self._sections,
        )

    def save(self, path: Path) -> None:
        path.write_text(self.html, encoding="utf-8")
